import datetime as dt
from typing import Optional, Union

import pydantic

from ..utils import File


def to_camel(snake_str: str) -> str:
    first, *others = snake_str.split("_")
    ret = "".join([first.lower(), *map(str.title, others)])
    return ret


class BaseAPIModel(pydantic.BaseModel):
    class Config:
        arbitrary_types_allowed = True
        alias_generator = to_camel
        allow_population_by_field_name = True

    def dict(self, **kwargs):
        """Returns camelCased version and removes any non specified optional fields """
        return super().dict(by_alias=True, exclude_unset=True)


class PatchAssetInput(BaseAPIModel):
    category: Optional[str]
    code: Optional[str]
    photo: Union[File, str, None]
    make: Optional[str]
    model: Optional[str]
    size: Optional[str]
    location: Optional[str]
    status: Optional[str]
    condition: Optional[str]
    installation_timestamp: Optional[dt.datetime]
    expected_life_years: Optional[dt.datetime]


class PatchRemarkInput(BaseAPIModel):
    resolved_timestamp: Optional[dt.datetime]
    description: Optional[str]
    severity: Optional[str]
    resolution: Optional[str]


class PatchServiceInput(BaseAPIModel):
    name: Optional[str]
    due_date: Optional[dt.date]
    performed_timestamp: Optional[dt.datetime]
    description: Optional[str]
    result: Optional[str]



from .base import API

route_to_tenant_query = """
  mutation routeToTenant($tenantUuid: UUID!, $routeCode: String!) {
    routeToTenant(input: {tenantUuid: $tenantUuid, routeCode: $routeCode}) {
      route {
        id
        uuid
      }
    }
  }
"""

route_to_asset_query = """
  mutation routeToAsset($assetUuid: UUID!, $routeCode: String!) {
    routeToAsset(input: {assetUuid: $assetUuid, routeCode: $routeCode}) {
      route {
        id
        uuid
      }
    }
  }
"""


class RoutingAPI(API):
    def route_to_tenant(self, route_code: str, tenant_uuid: str) -> str:
        result = self.perform_query(
            route_to_tenant_query, {"routeCode": route_code, "tenantUuid": tenant_uuid}
        )
        return result["routeToTenant"]["route"]["uuid"]

    def route_to_asset(self, route_code: str, asset_uuid: str) -> str:
        result = self.perform_query(
            route_to_asset_query, {"routeCode": route_code, "assetUuid": asset_uuid}
        )
        return result["routeToAsset"]["route"]["uuid"]



from .api.asset_api import AssetAPI
from .api.integration_api import IntegrationAPI
from .api.place_api import PlaceAPI
from .api.routing_api import RoutingAPI


class BSecureClient(PlaceAPI, AssetAPI, IntegrationAPI, RoutingAPI):
    DEFAULT_DOMAIN = "bsec.me"

    def __init__(self, domain=None, origin=None, encrypt=True, jwt=None):
        self.protocol = "https" if encrypt else "http"
        self.domain = domain or self.DEFAULT_DOMAIN
        self.origin = origin
        self.jwt = jwt

        # TBH this isn't ideal but at least we have proper typing for all
        # the apis now....
        # Ideally we just have PlaceAPI as an instansiated attribute on the client...
        # like BsecureClient().place.get_tenant_by_uid
        # This is still better than monkey patching this base class with methods dynamically
        for superclass in [PlaceAPI, AssetAPI, IntegrationAPI, RoutingAPI]:
            super(superclass, self).__init__(self)  # type: ignore



import datetime as dt
from typing import Dict, Optional

from ..utils import File, make_timestamp
from .base import API
from .types import PatchAssetInput, PatchRemarkInput, PatchServiceInput

remark_fields = """
    uuid
    createdTimestamp
    resolvedTimestamp
    description
    severity
    resolution
    created
    updated
"""

service_fields = """
    uuid
    name
    description
    createdTimestamp
    dueDate
    performedTimestamp
    result
    created
    updated
"""

asset_fields = """
    uuid
    code
    category
    make
    model
    size
    location
    status
    condition
    photo
    installationTimestamp
    expectedLifeYears
"""


push_asset_query = """
  mutation pushAsset(
    $tenantUuid: UUID!,
    $category: String!,
    $code: String!,
    $make: String,
    $model: String,
    $installationTimestamp: DateTime,
    $expectedLifeYears: Float,
    $photo: String
  ) {
    pushAsset(input: {
      tenantUuid: $tenantUuid,
      category: $category,
      code: $code,
      model: $model,
      make: $make,
      installationTimestamp: $installationTimestamp,
      expectedLifeYears: $expectedLifeYears,
      photo: $photo
    }) {
      asset {
        uuid
      }
    }
  }
"""

update_asset_query = """
  mutation updateAsset(
    $uuid: UUID!,
    $installationTimestamp: DateTime,
    $expectedLifeYears: Float,
    $photo: String
  ) {
    updateAsset(input: {
      uuid: $uuid,
      installationTimestamp: $installationTimestamp,
      expectedLifeYears: $expectedLifeYears,
      photo: $photo
    }) {
      asset {
        uuid
      }
    }
  }
"""

push_remark_query = """
  mutation pushRemark(
    $assetUuid: UUID!,
    $createdTimestamp: DateTime!,
    $resolvedTimestamp: DateTime,
    $description: String,
    $resolution: String
  ) {
    pushRemark(input: {
      assetUuid: $assetUuid,
      createdTimestamp: $createdTimestamp,
      resolvedTimestamp: $resolvedTimestamp,
      description: $description,
      resolution: $resolution
    }) {
      remark {
        uuid
      }
    }
  }
"""

push_service_query = """
  mutation pushService(
    $assetUuid: UUID!,
    $name: String!,
    $createdTimestamp: DateTime!,
    $description: String,
    $dueDate: Date,
    $performedTimestamp: DateTime,
    $result: String
  ) {
    pushService(input: {
      assetUuid: $assetUuid,
      name: $name,
      createdTimestamp: $createdTimestamp,
      description: $description,
      dueDate: $dueDate,
      performedTimestamp: $performedTimestamp,
      result: $result
    }) {
      service {
        uuid
      }
    }
  }
"""


patch_asset_mutation = f"""
  mutation patchAsset(
    $uuid: UUID!,
    $patch: PatchAssetInput!
  ) {{
    patchAsset(uuid: $uuid, patch: $patch) {{
      asset {{
        {asset_fields}
      }}
    }}
  }}
"""

patch_service_mutation = f"""
  mutation patchService(
    $uuid: UUID!,
    $patch: PatchServiceInput!
  ) {{
    patchService(uuid: $uuid, patch: $patch) {{
      service {{
          {service_fields}
      }}
    }}
  }}
"""

patch_remark_mutation = f"""
  mutation patchRemark(
    $uuid: UUID!,
    $patch: PatchRemarkInput!
  ) {{
    patchRemark(uuid: $uuid, patch: $patch) {{
      remark {{
          {remark_fields}
      }}
    }}
  }}
"""


class AssetAPI(API):
    def push_asset(
        self,
        tenant_uuid: str,
        category: str,
        code: Optional[str] = None,
        make: Optional[str] = None,
        model: Optional[str] = None,
        installation_timestamp: Optional[dt.datetime] = None,
        expected_life_years: Optional[float] = None,
        photo: Optional[File] = None,
    ) -> str:
        photo_file_id = self.upload_file(photo)
        response_data = self.perform_query(
            push_asset_query,
            self.make_variables(
                tenantUuid=tenant_uuid,
                category=category,
                code=code,
                make=make,
                model=model,
                installationTimestamp=make_timestamp(installation_timestamp),
                expectedLifeYears=expected_life_years,
                photo=photo_file_id,
            ),
        )
        return response_data["pushAsset"]["asset"]["uuid"]

    def update_asset(
        self,
        uuid: str,
        installation_timestamp: Optional[dt.datetime] = None,
        expected_life_years: Optional[float] = None,
        photo: Optional[File] = None,
    ) -> str:
        photo_file_id = self.upload_file(photo)
        response_data = self.perform_query(
            update_asset_query,
            self.make_variables(
                uuid=uuid,
                installationTimestamp=make_timestamp(installation_timestamp),
                expectedLifeYears=expected_life_years,
                photo=photo_file_id,
            ),
        )
        return response_data["updateAsset"]["asset"]["uuid"]

    def push_asset_remark(
        self,
        asset_uuid: str,
        created_timestamp: dt.datetime,
        resolved_timestamp: Optional[dt.datetime] = None,
        description: Optional[str] = None,
        resolution: Optional[str] = None,
    ) -> str:
        response_data = self.perform_query(
            push_remark_query,
            self.make_variables(
                assetUuid=asset_uuid,
                createdTimestamp=make_timestamp(created_timestamp),
                resolvedTimestamp=make_timestamp(resolved_timestamp),
                description=description,
            ),
        )
        return response_data["pushRemark"]["remark"]["uuid"]

    def push_asset_service(
        self,
        asset_uuid: str,
        name: str,
        created_timestamp: dt.datetime,
        description: Optional[str] = None,
        due_date: Optional[dt.date] = None,
        performed_timestamp: Optional[dt.datetime] = None,
        result: Optional[str] = None,
    ) -> str:
        response_data = self.perform_query(
            push_service_query,
            self.make_variables(
                assetUuid=asset_uuid,
                name=name,
                description=description,
                createdTimestamp=make_timestamp(created_timestamp),
                dueDate=make_timestamp(due_date),
                performedTimestamp=make_timestamp(performed_timestamp),
                result=result,
            ),
        )
        return response_data["pushService"]["service"]["uuid"]

    def patch_asset(
        self,
        asset_uuid: str,
        patch: PatchAssetInput,
    ) -> dict:
        photo = patch.photo
        if isinstance(photo, File):
            patch.photo = self.upload_file(photo)
        response_data = self.perform_query(
            patch_asset_mutation, variables={"uuid": asset_uuid, "patch": patch.dict()}
        )
        return response_data["patchAsset"]["asset"]

    def patch_service(self, service_uuid: str, patch: PatchServiceInput) -> Dict:
        response_data = self.perform_query(
            patch_service_mutation,
            variables={"uuid": service_uuid, "patch": patch.dict()},
        )
        return response_data["patchService"]["service"]

    def patch_remark(self, remark_uuid: str, patch: PatchRemarkInput) -> Dict:
        response_data = self.perform_query(
            patch_remark_mutation,
            variables={"uuid": remark_uuid, "patch": patch.dict()},
        )
        return response_data["patchRemark"]["remark"]



from typing import Optional

import requests

from ..exceptions import QueryException
from ..utils import File, make_serializeable, remove_trailing_slash

upload_file_mutation = """
  mutation uploadFile($filename: String!) {
    uploadFile(filename: $filename) {
      guid,
      presignedUrl
    }
  }
"""


class API:
    SUCCESS_RESPONSE_CODES = (200, 201)

    def __init__(self, client):
        self.client = client

    def perform_query(self, query, variables=None, headers=None):
        response = self.send_query(query, variables, headers)
        response.raise_for_status()
        response_data = response.json()
        self.check_for_errors(response_data)
        return response_data["data"]

    def get_url(self) -> str:
        domain: str = remove_trailing_slash(self.client.domain)
        if domain.startswith("https://") or domain.startswith("http://"):
            return f"{domain}/graphql/"
        else:
            return f"{self.client.protocol}://{domain}/graphql/"

    def send_query(self, query, variables, headers):
        auth_headers = {}
        url = self.get_url()

        if self.client.jwt:
            auth_headers["Authorization"] = f"Bearer {self.client.jwt}"
        response = requests.post(
            url,
            json={"query": query, "variables": make_serializeable(variables)},
            headers={
                **({"Origin": self.client.origin} if self.client.origin else {}),
                **auth_headers,
                **(headers or {}),
            },
        )
        # This will only trigger for 400/500 django view exceptions.
        if response.status_code not in self.SUCCESS_RESPONSE_CODES:
            msg = response.content
            raise QueryException(f"BSECURE error: {str(msg)}")

        return response

    def upload_file(self, file: Optional[File]) -> Optional[str]:
        """Uploads a `File` in echange for a tmpfile_uuid

        This is used by other mutations to upload a file to a temporary s3 location
        before it is published with another mutation
        """
        if not file:
            return None
        response_data = self.perform_query(upload_file_mutation, {"filename": file.filename})
        if "uploadFile" not in response_data.keys() or not response_data["uploadFile"]:
            raise QueryException("Failed to upload file: You do not have permission")

        tmpfile_uuid = response_data["uploadFile"]["guid"]
        presigned_url = response_data["uploadFile"]["presignedUrl"]

        # Upload file to s3
        upload_response = requests.put(presigned_url, data=file.content.read())

        if upload_response.status_code != 200:
            raise QueryException(f"Failed to upload file: {upload_response.text}")

        return tmpfile_uuid

    def check_for_errors(self, response_data):
        # GraphQL responses only throw 400s and 500s if something goes
        # very wrong, so we need to check the error fields in addition.
        if response_data.get("errors"):
            try:
                msg = response_data["errors"][0]["message"]
            except Exception:
                msg = "unknown reason"
            raise QueryException(f"BSECURE error: {msg}")

    def make_variables(self, **kwargs):
        return {key: value for key, value in kwargs.items() if value is not None}



from typing import Dict, List, Optional, Tuple

from ..utils import File
from .base import API

get_tenant_query = """
  query getTenant(
    $uuid: UUID!
  ) {
    tenant(uuid: $uuid) {
      id
      uuid
      name
      clientName
      source
      place {
        id
        uuid
        name
        shortName
        description
        formattedAddress
        gcsCoord
        gnafId
      }
    }
  }
"""

push_tenant_query = """
  mutation pushTenant(
    $name: String!,
    $formattedAddress: String!,
    $latitude: Float!,
    $longitude: Float!,
    $source: String!,
    $gnafId: String,
    $clientName: String,
    $frontPhoto: String
  ) {
    pushTenant(input: {
      name: $name,
      formattedAddress: $formattedAddress,
      latitude: $latitude,
      longitude: $longitude,
      source: $source,
      gnafId: $gnafId,
      clientName: $clientName,
      frontPhoto: $frontPhoto
    }) {
      tenant {
        uuid
      }
    }
  }
"""

all_tenant_documents_query = """
  query tenantDocuments($uuid: UUID!) {
    tenant(uuid: $uuid) {
      id
      uuid
      name
      tenantdocumentSet {
        edges {
          node {
            id
            uuid
            title
            uploadedTimestamp
            filename
          }
        }
      }
    }
  }
"""

push_tenant_photo_mutation = """
  mutation pushTenantPhoto(
    $tenant_uuid: UUID!,
    $tmpfile_uuid: UUID!,
  ) {
    pushTenantPhoto(
      tenantUuid: $tenant_uuid,
      tmpfileUuid: $tmpfile_uuid,
    ) {
      tenantPhoto {
        uuid
      }
    }
  }
"""

push_tenant_document_mutation = """
  mutation pushTenantDocument(
    $tenant_uuid: UUID!,
    $title: String!,
    $tmpfile_uuid: UUID!,
  ) {
    pushTenantDocument(
      tenantUuid: $tenant_uuid,
      title: $title,
      tmpfileUuid: $tmpfile_uuid,
    ) {
      tenantDocument {
        uuid
      }
    }
  }
"""


class PlaceAPI(API):
    # TODO: Would be nice to be able to share these with the server
    # implementation.
    PUBLIC = "Public"
    AUDIT = "Audit"
    SERVICE = "Service"

    def push_tenant(
        self,
        name: str,
        formatted_address: str,
        gcs_coord: Tuple[float, float],
        source: str,
        gnaf_id: Optional[str] = None,
        client_name: Optional[str] = None,
        front_photo: Optional[File] = None,
    ) -> str:
        """Pushes a tenant into bsecure

        Creates a tenant for a place (and a plce if it does not exist).

        An existing place is chosen if:
            the gnaf_id matches
        OR
            the formatted_address and coordinates are close

        The source should also be provided eg: workforce or logbooks
        """
        response_data = self.perform_query(
            push_tenant_query,
            self.make_variables(
                name=name,
                formattedAddress=formatted_address,
                longitude=gcs_coord[0],
                latitude=gcs_coord[1],
                gnafId=gnaf_id,
                clientName=client_name,
                source=source.lower(),
            ),
        )
        tenant_uuid = response_data["pushTenant"]["tenant"]["uuid"]
        if front_photo:
            self.push_tenant_photo(tenant_uuid, front_photo)
        return tenant_uuid

    def all_tenant_documents(self, tenant_uuid: str) -> List[Dict]:
        response_data = self.perform_query(all_tenant_documents_query, {"uuid": tenant_uuid})
        return [edge["node"] for edge in response_data["tenant"]["tenantdocumentSet"]["edges"]]

    def get_tenant(self, uuid: str) -> Dict:
        """Returns a tenant (and parent place) by uuid"""
        response_data = self.perform_query(get_tenant_query, {"uuid": uuid})
        return response_data["tenant"]

    def push_tenant_photo(self, tenant_uuid: str, photo: File) -> str:
        assert photo
        tmpfile_uuid = self.upload_file(photo)
        response_data = self.perform_query(
            push_tenant_photo_mutation,
            {
                "tenant_uuid": tenant_uuid,
                "tmpfile_uuid": tmpfile_uuid,
            },
        )
        return response_data["pushTenantPhoto"]["tenantPhoto"]["uuid"]

    def push_tenant_document(self, tenant_uuid: str, title: str, document: File) -> str:
        assert document
        tmpfile_uuid = self.upload_file(document)
        response_data = self.perform_query(
            push_tenant_document_mutation,
            {
                "tenant_uuid": tenant_uuid,
                "title": title,
                "tmpfile_uuid": tmpfile_uuid,
            },
        )
        return response_data["pushTenantDocument"]["tenantDocument"]["uuid"]



import datetime as dt
import enum
from copy import deepcopy
from io import BytesIO
from typing import Any, Dict, Optional, Union
from uuid import UUID


class File:
    """Bsecure `File` type"""

    def __init__(self, content: BytesIO, filename: Optional[str] = None):
        self.filename = filename or content.name
        self.content = content


class FileID:
    def __init__(self, id):
        self.id = id

    def __str__(self):
        return self.id


def remove_trailing_slash(path: str) -> str:
    if path and path[-1] == "/":
        path = path[:-1]
    return path


def make_timestamp(value: Optional[Union[dt.datetime, dt.date]]) -> Optional[str]:
    if isinstance(
        value,
        (
            dt.datetime,
            dt.date,
        ),
    ):
        return value.isoformat()
    else:
        return None


def to_serializable(val: Any):
    """JSON serializer for objects not serializable by default"""
    if isinstance(val, (dt.datetime, dt.date, dt.time)):
        return val.isoformat()
    elif isinstance(val, enum.Enum):
        return val.value
    elif isinstance(val, UUID):
        return str(val)
    elif hasattr(val, "__dict__"):
        return val.__dict__
    else:
        return val


def make_serializeable(dictionary: Dict) -> Dict:
    new_dict = deepcopy(dictionary)

    for k, v in new_dict.items():
        if isinstance(v, dict):
            new_dict[k] = make_serializeable(v)
        else:
            new_dict[k] = to_serializable(v)
    return new_dict



start_integration_query = """
  mutation startIntegration($endpoint: String!, $host: String!) {
    startIntegration(input: {endpoint: $endpoint, host: $host}) {
      token
    }
  }
"""

finish_integration_query = """
  mutation finishIntegration($host: String!) {
    finishIntegration(input: {host: $host}) {
      key
    }
  }
"""

push_property_query = """
  mutation pushProperty(
    $name: String!,
    $formatted: String!,
    $latitude: BigFloat!,
    $longitude: BigFloat!,
    $client: String!
  ) {
    pushProperty(input: {
      name: $name,
      formatted: $formatted,
      latitude: $latitude,
      longitude: $longitude,
      client: $client
    }) {
      uuid
    }
  }
"""

push_asset_query = """
  mutation pushAsset(
    $tenantGuid: UUID!,
    $category: String!,
    $code: String!,
    $make: String,
    $model: String,
    $photo: String
  ) {
    pushAsset(input: {
      tenantGuid: $tenantGuid,
      category: $category,
      code: $code,
      model: $model,
      make: $make,
      photo: $photo
    }) {
      asset {
        guid
      }
    }
  }
"""

push_remark_query = """
  mutation pushRemark(
    $assetGuid: UUID!,
    $createdTime: Datetime!,
    $resolvedTime: Datetime,
    $description: String,
    $resolution: String
  ) {
    pushRemark(input: {
      assetGuid: $assetGuid,
      createdTime: $createdTime,
      resolvedTime: $resolvedTime,
      description: $description,
      resolution: $resolution
    }) {
      uuid
    }
  }
"""

push_service_query = """
  mutation pushService(
    $assetGuid: UUID!,
    $name: String!,
    $description: String,
    $createdTime: Datetime!,
    $dueTime: Date,
    $performedTime: Datetime,
    $result: String
  ) {
    pushService(input: {
      assetGuid: $assetGuid,
      name: $name,
      description: $description,
      createdTime: $createdTime,
      dueTime: $dueTime,
      performedTime: $performedTime,
      result: $result
    }) {
      service {
        guid
      }
    }
  }
"""

route_to_place_query = """
  mutation routeToPlace($tenantGuid: UUID!, $routeGuid: String!) {
    routeToPlace(input: {tenantGuid: $tenantGuid, routeGuid: $routeGuid}) {
      clientMutationId
    }
  }
"""

route_to_asset_query = """
  mutation routeToAsset($assetGuid: UUID!, $routeGuid: String!) {
    routeToAsset(input: {assetGuid: $assetGuid, routeGuid: $routeGuid}) {
      clientMutationId
    }
  }
"""

upload_place_document_query = """
  mutation uploadPlaceDocument($filename: String!) {
    uploadPlaceDocumentFilename(input: {filename: $filename}) {
      filename,
      presignedUrl
    }
  }
"""

upload_asset_photo_query = """
  mutation uploadAssetPhoto($filename: String!) {
    uploadAssetPhoto(input: {filename: $filename}) {
      filename,
      presignedUrl
    }
  }
"""

all_place_documents_by_tenant_query = """
  query allPlaceDocumentsByTenant($tenantGuid: UUID!) {
    allPlaceDocumentsByTenant(tenantGuid: $tenantGuid) {
      nodes {
        title
      }
    }
  }
"""



import logging
from typing import Callable

from .base import API

logger = logging.getLogger(__name__)


start_integration_query = """
  mutation startIntegration($endpoint: String!, $host: String!, $secret: String!) {
    startIntegration(input: {endpoint: $endpoint, host: $host, secret: $secret}) {
      token
    }
  }
"""

finish_integration_query = """
  mutation finishIntegration($host: String!) {
    finishIntegration(input: {host: $host}) {
      jwt
    }
  }
"""

initiate_integration_mutation = """
  mutation initiateIntegration($host: String!, $organisation_name: String!, $bsecure_integration_secret: String!) {
      initiateIntegration(input: {
          host: $host, organisationName: $organisation_name, bsecureIntegrationSecret: $bsecure_integration_secret
          }) {
          jwt
      }
  }
"""


class IntegrationAPI(API):
    def integrate(self, prepare_endpoint: Callable, endpoint: str, host: str, secret: str) -> str:
        logging.warn("This is a deprecated endpoint please use initiate_integration")
        response_data = self.perform_query(
            start_integration_query,
            {"endpoint": endpoint, "host": host, "secret": secret},
        )
        token = response_data["startIntegration"]["token"]
        prepare_endpoint(token)
        response_data = self.perform_query(finish_integration_query, {"host": host})
        if not response_data["finishIntegration"]:
            response_data["finishIntegration"] = {"jwt": None}
        return response_data["finishIntegration"]["jwt"]

    def initiate_integration(
        self, host: str, organisation_name: str, bsecure_integration_secret: str
    ) -> str:
        """Integrate this host with BSecure. Returns a JWT token if successful."""
        response_data = self.perform_query(
            initiate_integration_mutation,
            {
                "organisation_name": organisation_name,
                "host": host,
                "bsecure_integration_secret": bsecure_integration_secret,
            },
        )
        jwt = response_data["initiateIntegration"]["jwt"]
        return jwt
